#include <stdio.h>
#include <time.h>
#include <string.h>

#include <gtk/gtk.h>

#define GTK_EVENT_ACTIVATE "activate"
#define GTK_EVENT_DESTROY "destroy"
#define GTK_EVENT_CLICKED "clicked"
#define GTK_EVENT_CHANGED "changed"

#define DEBUG

/* GtkWidgets */
static GtkWidget *entry_date;
static GtkWidget *entry_hour;
static GtkWidget *entry_minute;
static GtkWidget *entry_second;

/* Time info buffer */
static time_t time_now;
static struct tm *tm_now;
static char active_date[11];
static char active_hour[3];
static char active_min[3];
static char active_sec[3];

/* Timezone info buffer */
#define MAX_TIME_STATUS_LINE 8
static char timezone_cur[32] = {0};
static char timezone_new[32] = {0};
static char timezone_buffer[512][32];
static int timezone_num = 0;

static void date_init(GtkEntry *entry_date, GtkEntry *entry_hour, GtkEntry *entry_minute, GtkEntry *entry_second);

static void update_date_time() {
    date_init(GTK_ENTRY(entry_date), GTK_ENTRY(entry_hour),
            GTK_ENTRY(entry_minute), GTK_ENTRY(entry_second));
}

static int get_timezone(const char *src, char *timezone) {
    const char *buf = src;
    int index = 0;

    /* Skip space char */
    while(buf[index] == ' ') index++;

    for (int i = 0; ; i++, index++) {
        if (buf[index] != ' ') {
            timezone[i] = buf[index];
        } else {
            /* End by space */
            timezone[i] = 0;
            break;
        }
    }

    return 0;
}

static void set_timedatectl(char *command, char *data) {
    char cmd[64] = "timedatectl ";
    strcat(cmd, command);
    strcat(cmd, " ");
    strcat(cmd, data);
#ifdef DEBUG
    g_print("[DEBUG] CMD: %s\n", cmd);
#endif

    FILE *fp = popen(cmd, "r");
    pclose(fp);
}

static void set_timezone(char *timezone) {
    if (strcmp(timezone_cur, timezone)) {
        set_timedatectl("set-timezone", timezone);
        /* Update current timezone to be set */
        strcpy(timezone_cur, timezone);
        /* Update current date time after setting timezone */
        update_date_time();
    }
}

static void set_ntp(char *enable) {
    set_timedatectl("set-ntp", enable);
}

static void set_date_time(const char *date, const char *hour, const char *min, const char *sec) {
    char data[32] = {0};

    if (strcmp(active_date, date)) {
        sprintf(data, "'%s %s:%s:%s'", date, hour, min, sec);
        set_ntp("no"); // set ntp as no before seting time/date
        set_timedatectl("set-time", data);
        strcpy(active_date, date);
        strcpy(active_hour, hour);
        strcpy(active_min, min);
        strcpy(active_sec, sec);
    } else if (strcmp(active_hour, hour) || strcmp(active_min, min) || strcmp(active_sec, sec)) {
        sprintf(data, "'%s:%s:%s'", hour, min, sec);
        set_ntp("no"); // set ntp as no before seting time/date
        set_timedatectl("set-time", data);
        strcpy(active_hour, hour);
        strcpy(active_min, min);
        strcpy(active_sec, sec);
    }
}

static void get_time_date_status() {
    char time_status[8][32];
    FILE *fp = popen("timedatectl status | grep \"Time zone\"", "r");

    for (int i = 0; i < MAX_TIME_STATUS_LINE; i++) {
        char *buf = fgets(time_status[i], 32, fp);
        if (buf == NULL) {
            break;
        } else if (strstr(time_status[i], "Time zone") != NULL) {
            int line_size = strlen(buf);
            fseek(fp, line_size, SEEK_CUR);

            buf = strchr(time_status[i], ':');
            if (buf != NULL) {
                get_timezone(buf + 1, timezone_cur);
                if (!strcmp(timezone_cur, "Universal")) {
                    strcpy(timezone_cur, "UTC"); // Set default "Universal" timezone as UTC.
                }
                strcpy(timezone_new, timezone_cur);
            }
        }
    }
#ifdef DEBUG
    g_print("[DEBUG] Current timezone: %s\n", timezone_cur);
#endif
    pclose(fp);
}

static void combo_handler(GObject *combo, gpointer user_data) {
    GtkTreeIter  iter;
    GtkTreeModel *model;
    gchar        *string = NULL;

    if (!g_strcmp0((gchar *)user_data, "timezone")) {
        /* Obtain currently selected item from combo box.
         * If nothing is selected, do nothing. */
        if (gtk_combo_box_get_active_iter(GTK_COMBO_BOX(combo), &iter)) {
            /* Obtain data model from combo box. */
            model = gtk_combo_box_get_model(GTK_COMBO_BOX(combo));

            /* Obtain string from model. */
            gtk_tree_model_get(model, &iter, 0, &string, -1);
        }
        /* Update current timezone */
        strcpy(timezone_new, ( string ? string : "NULL" ));

        /* Free string (if not NULL). */
        if (string)
            g_free(string);
            
#ifdef DEBUG
        /* Print string to the console - if string is NULL, print NULL. */
        g_print("[DEBUG] Selected (complex): >> %s <<\n", timezone_new);
#endif
    }
}

static void timezone_init(GtkWidget *window_vbox) {
    GtkWidget *combo_timezone;
    GtkListStore *store = NULL;
    GtkTreeIter iter;
    GtkTreeIter active_iter;
    GtkCellRenderer *cell = NULL;

    /* Timezone */
    GtkWidget *align_timezone = gtk_alignment_new(0.5,0.5,0,0);;
    GtkWidget *timezone_hbox = gtk_hbox_new(TRUE, 2);

    GtkWidget *label_timezone = gtk_label_new("TimeZone:");
    gtk_box_pack_start(GTK_BOX(timezone_hbox), label_timezone, FALSE, FALSE, 3);

    combo_timezone = gtk_combo_box_new();
    gtk_box_pack_start(GTK_BOX(timezone_hbox), combo_timezone, FALSE, FALSE, 3);

    gtk_container_add(GTK_CONTAINER(align_timezone), timezone_hbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_timezone), FALSE, FALSE, 5);

    FILE *fp = popen("timedatectl list-timezones", "r");

    timezone_num = 0;
    for (int i = 0; i < 512; i++) {
        char *buf = fgets(timezone_buffer[i], 32, fp);
        if (buf == NULL) {
            break;
        } else {
            int size = strlen(timezone_buffer[i]);
            timezone_buffer[i][size - 1] = 0; // remove '\n' char.
            fseek(fp, strlen(buf), SEEK_CUR);
            timezone_num++;
        }
    }

    pclose(fp);

#ifdef DEBUG
    g_print("[DEBUG] Get %d timezones\n", timezone_num);
#endif

    store = gtk_list_store_new(1, G_TYPE_STRING);

    for (int i = 0; i < timezone_num; i++) {
        gtk_list_store_append(store, &iter);
        if (!g_strcmp0(timezone_buffer[i], timezone_cur)) {
            active_iter = iter;
        }
        gtk_list_store_set(store, &iter, 0, timezone_buffer[i], -1);
    }

    gtk_combo_box_set_model(GTK_COMBO_BOX(combo_timezone), GTK_TREE_MODEL(store));
    
    cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo_timezone), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo_timezone), cell, "text", 0, NULL);

    /* Set default active item by iter */
    gtk_combo_box_set_active_iter(GTK_COMBO_BOX(combo_timezone), &active_iter);

    g_signal_connect(combo_timezone, GTK_EVENT_CHANGED, G_CALLBACK(combo_handler), "timezone");
}

static void combobox_timezone_changed_cb(GObject *object, gpointer user_data) {
#ifdef DEBUG
    g_print("combobox_timezone_changed_cb");
#endif
}

static void button_handler(GObject *button, gpointer user_data) {
    const gchar *date;
    const gchar *hour;
    const gchar *min;
    const gchar *sec;

    if (!g_strcmp0((gchar *)user_data, "ok")) {
        date = gtk_entry_get_text(GTK_ENTRY(entry_date));
        hour = gtk_entry_get_text(GTK_ENTRY(entry_hour));
        min = gtk_entry_get_text(GTK_ENTRY(entry_minute));
        sec = gtk_entry_get_text(GTK_ENTRY(entry_second));
        set_date_time(date, hour, min, sec);
        set_timezone(timezone_new);
    } else if (!g_strcmp0((gchar *)user_data, "cancel")) {
        gtk_main_quit();
    }
}

static void date_init(GtkEntry *entry_date, GtkEntry *entry_hour, GtkEntry *entry_minute, GtkEntry *entry_second) {
    /* Get current local time */
    time(&time_now);
    tm_now = localtime(&time_now);

    sprintf(active_date, "%04d-%02d-%02d", (1900 + tm_now->tm_year), (tm_now->tm_mon + 1), tm_now->tm_mday);
#ifdef DEBUG
    g_print("[DEBUG] Date: %s\n", active_date);
#endif

    sprintf(active_hour, "%02d", tm_now->tm_hour);
    sprintf(active_min, "%02d", tm_now->tm_min);
    sprintf(active_sec, "%02d", tm_now->tm_sec);
#ifdef DEBUG
    g_print("[DEBUG] Time: %s:%s:%s\n", active_hour, active_min, active_sec);
#endif

    gtk_entry_set_text(entry_date, active_date);
    gtk_entry_set_text(entry_hour, active_hour);
    gtk_entry_set_text(entry_minute, active_min);
    gtk_entry_set_text(entry_second, active_sec);
}

static int date_widget_init(GtkWidget *window_vbox) {
    GtkWidget *align_date;
    GtkWidget *align_hour;
    GtkWidget *align_minute;
    GtkWidget *align_second;
    GtkWidget *combo_hour;
    GtkWidget *combo_minute;
    GtkWidget *combo_second;
    gchar text[3] = {0};

    align_date = gtk_alignment_new(0.5,0.5,0,0);
    align_hour = gtk_alignment_new(0.5,0.5,0,0);
    align_minute = gtk_alignment_new(0.5,0.5,0,0);
    align_second = gtk_alignment_new(0.5,0.5,0,0);

    /* Date */
    GtkWidget *date_hbox = gtk_hbox_new(TRUE, 2);

    GtkWidget *label_date = gtk_label_new("Date:");
    gtk_box_pack_start(GTK_BOX(date_hbox), label_date, FALSE, FALSE, 3);

    entry_date = gtk_entry_new();
    if (entry_date == NULL) {
        g_print("[ERROR] create entry date object failed.\n");
        return -1;
    }
    gtk_box_pack_start(GTK_BOX(date_hbox), entry_date, FALSE, FALSE, 3);

    gtk_container_add(GTK_CONTAINER(align_date), date_hbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_date), FALSE, FALSE, 5);

    /* Time */
    GtkWidget *hour_hbox = gtk_hbox_new(TRUE, 2);
    GtkWidget *minute_hbox = gtk_hbox_new(TRUE, 2);
    GtkWidget *second_hbox = gtk_hbox_new(TRUE, 2);

    GtkWidget *label_time = gtk_label_new("Time:");
    gtk_box_pack_start(GTK_BOX(hour_hbox), label_time, FALSE, FALSE, 3);

    /* Hour */
    GtkWidget *label_hour = gtk_label_new("H:");
    gtk_box_pack_start(GTK_BOX(hour_hbox), label_hour, FALSE, FALSE, 3);

    combo_hour = gtk_combo_box_text_new_with_entry();
    /* Add hour combo box context */
    for (int i = 0; i < 24; i++) {
        sprintf(text, "%d", i);
        gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(combo_hour), text);
    }
    entry_hour = (GTK_BIN(combo_hour)->child);
    gtk_box_pack_start(GTK_BOX(hour_hbox), combo_hour, FALSE, FALSE, 3);

    gtk_container_add(GTK_CONTAINER(align_hour), hour_hbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_hour), FALSE, FALSE, 5);

    /* Minute */
    GtkWidget *label_blank1 = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(minute_hbox), label_blank1, FALSE, FALSE, 3);

    GtkWidget *label_minute = gtk_label_new("M:");
    gtk_label_set_justify(GTK_LABEL(label_minute), GTK_JUSTIFY_RIGHT);
    gtk_box_pack_start(GTK_BOX(minute_hbox), label_minute, FALSE, FALSE, 3);

    combo_minute = gtk_combo_box_text_new_with_entry();
    /* Add minute combo box context */
    for (int i = 0; i < 60; i++) {
        sprintf(text, "%d", i);
        gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(combo_minute), text);
    }
    entry_minute = (GTK_BIN(combo_minute)->child);
    gtk_box_pack_start(GTK_BOX(minute_hbox), combo_minute, FALSE, FALSE, 3);

    gtk_container_add(GTK_CONTAINER(align_minute), minute_hbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_minute), FALSE, FALSE, 5);

    /* Second */
    GtkWidget *label_blank2 = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(second_hbox), label_blank2, FALSE, FALSE, 3);

    GtkWidget *label_second = gtk_label_new("S:");
    gtk_label_set_justify(GTK_LABEL(label_second), GTK_JUSTIFY_RIGHT);
    gtk_box_pack_start(GTK_BOX(second_hbox), label_second, FALSE, FALSE, 3);

    combo_second = gtk_combo_box_text_new_with_entry();
    /* Add second combo box context */
    for (int i = 0; i < 60; i++) {
        sprintf(text, "%d", i);
        gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(combo_second), text);
    }
    entry_second = (GTK_BIN(combo_second)->child);
    gtk_box_pack_start(GTK_BOX(second_hbox), combo_second, FALSE, FALSE, 3);

    gtk_container_add(GTK_CONTAINER(align_second), second_hbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_second), FALSE, FALSE, 5);

    date_init(GTK_ENTRY(entry_date), GTK_ENTRY(entry_hour), GTK_ENTRY(entry_minute), GTK_ENTRY(entry_second));

    // /* Set default time combo box item */
    // gtk_combo_box_set_active_id(GTK_COMBO_BOX(combo_hour), active_hour);
    // gtk_combo_box_set_active_id(GTK_COMBO_BOX(combo_minute), active_min);
    // gtk_combo_box_set_active_id(GTK_COMBO_BOX(combo_second), active_sec);
    // gtk_combo_box_set_active(GTK_COMBO_BOX(combo_hour), tm_now->tm_hour);

    return 0;
}

// void close_window(GtkWidget *dialog, GtkWidget *window){
//     gint result = gtk_dialog_get_response_for_widget(GTK_DIALOG(dialog), window);
//     g_print("[DEBUG] result is %d\n", result);
//     /* Close window without exit program*/
//     //gtk_main_quit();
//     gtk_widget_destroy (GTK_WIDGET (dialog));
//     gtk_widget_destroy (GTK_WIDGET (window));
// }

int main(int argc, char **argv) {
    GtkWidget *window;
    GtkWidget *align_window;
    
    gtk_init(&argc, &argv);

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    align_window = gtk_alignment_new(0.5,0.5,0,0);

    gtk_window_set_title(GTK_WINDOW(window), "Date & Time");
    gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
    gtk_window_set_default_size(GTK_WINDOW(window), 480, 240);

    /* Create vbox for whole style(vertical) */
    GtkWidget *window_vbox = gtk_vbox_new(TRUE, 2);
    GtkWidget *date_time_vbox = gtk_vbox_new(TRUE, 2);

    /* Date */
    date_widget_init(date_time_vbox);

    /* Timezone */
    get_time_date_status();
    timezone_init(date_time_vbox);

    gtk_container_add(GTK_CONTAINER(align_window), date_time_vbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_window), FALSE, FALSE, 0);

    /* OK and Cancel */
    GtkWidget *align_button = gtk_alignment_new(0.5,0.5,0,0);
    GtkWidget *button_hbox = gtk_hbox_new(TRUE, 2);

    GtkWidget *button_ok = gtk_button_new_with_label("Ok");
    GtkWidget *button_cancel = gtk_button_new_with_label("Cancel");

    gtk_widget_set_size_request(button_ok, 120, 60);
    gtk_widget_set_size_request(button_cancel,  120, 60);
    //gtk_widget_set_size_request(button_hbox, 240, 80);

    gtk_box_pack_start(GTK_BOX(button_hbox), button_ok, FALSE, FALSE, 50);
    gtk_box_pack_start(GTK_BOX(button_hbox), button_cancel, FALSE, FALSE, 50);

    gtk_container_add(GTK_CONTAINER(align_button), button_hbox);
    gtk_box_pack_start(GTK_BOX(window_vbox), GTK_WIDGET(align_button), FALSE, FALSE, 2);

    g_signal_connect(button_ok, GTK_EVENT_CLICKED, G_CALLBACK(button_handler), "ok");
    g_signal_connect(button_cancel, GTK_EVENT_CLICKED, G_CALLBACK(button_handler), "cancel");

    /* Add to container */
    gtk_container_add(GTK_CONTAINER(window), window_vbox);

    g_signal_connect(window, GTK_EVENT_DESTROY, G_CALLBACK(gtk_main_quit), NULL);

    gtk_widget_show_all(window);
    gtk_main();

    // gint result = gtk_dialog_run (GTK_DIALOG (dialog));
    // switch (result) {
    //     case GTK_RESPONSE_REJECT:
    //        break;
    //     case GTK_RESPONSE_OK:
    //        //
    //        break;
    //     default:
    //        break;
    // }
    // gtk_widget_destroy (GTK_WIDGET (dialog));

    return 0;
}
